#include <ros/ros.h>
#include <geometry_msgs/PoseArray.h>
#include <geometry_msgs/Pose.h>
#include <std_msgs/Float32.h>
#include <std_msgs/String.h>
#include <geometry_msgs/PoseWithCovarianceStamped.h>
#include <fstream>
#include <string>

class RobotManager{
    private:
        ros::NodeHandle nh;
	ros::Publisher robotpose_pub;//キャリブレーション用の位置
	ros::Publisher calibration_pub;//キャリブレーション用のメッセージ
	ros::Subscriber amcl_sub;//AMCL_POS受け取り用
	ros::Subscriber robotcamerapose_sub;//ロボット推定位置受け取り用
        ros::Subscriber robotmsg_sub;//ロボットメッセージ受け取り用
	//変数
        geometry_msgs::Pose robot_camerapose;	//*ロボット位置推定用の変数
        //geometry_msgs::PoseWithCovarianceStamped robot_selfpose;	/*自己位置位置用の変数*/
        geometry_msgs::Pose robot_selfpose;	/*自己位置位置用の変数*/
        std_msgs::String robot_msg;        /*ロボットのメッセージの変数*/
        float distance;	/*距離の変数*/
        float threshold_distance;	/*距離の閾値の変数*/
        bool caribration_flag;
        bool amcl_flag,camerapose_flag;
        int count;
        std::ofstream writing_file;
    public:
        RobotManager();
	
        void robotposePub(void);
        void calibrationPub(void);
        void robotmsgCallback(const std_msgs::String message);
        void amclCallback(const geometry_msgs::PoseWithCovarianceStamped& msg);
        void poseCallback(const geometry_msgs::Pose pose);

        void comparePostion(void);
        void mainloop(void);
        void setup(void);
        void robotmsgcheck(void);
        void dummy(void);
};

RobotManager::RobotManager()
{
    //キャリブレーション用の位置
    robotpose_pub = nh.advertise<geometry_msgs::Pose>("/tb3_01/calibration_pose", 5);
    //キャリブレーション用のメッセージ
    calibration_pub = nh.advertise<std_msgs::String>("/tb3_01/manager_messege", 5);
    
    //AMCL_POS受け取り用
    amcl_sub = nh.subscribe("/tb3_01/amcl_pose", 1, &RobotManager::amclCallback, this);
    //AMCL_POS受け取り用
    robotcamerapose_sub=nh.subscribe("/robot_camepose", 1, &RobotManager::poseCallback, this);
    //ロボットメッセージ受け取り用
    robotmsg_sub = nh.subscribe("/robot_messege", 1, &RobotManager::robotmsgCallback, this);

}
void RobotManager::setup()
{
    ROS_INFO("setup");
    std::vector<double> robot_pose,camera_pose;
    //ros::param::get("robot_pose", robot_pose);  
    //ros::param::get("camera_pose",camera_pose);  
    //ros::param::get("/distance",distance);  
    ros::param::get("threshold_distance",threshold_distance);
    ROS_INFO("threshold_distance %f",threshold_distance);  
    caribration_flag = false;
    distance = 0;
    amcl_flag=false;
    camerapose_flag=false;
    //robot_camerapose.position.x = camera_pose[0];
    //robot_camerapose.position.y =camera_pose[1];
    //robot_selfpose.position.x =robot_pose[0];
    //robot_selfpose.position.y =robot_pose[1];
    count=0;
}
void RobotManager::dummy()
{
    //ROS_INFO("dummy %f",robot_camerapose.position.x);
    int limit=1000;
    if(count%limit==0){
      robot_camerapose.position.x +=0.1 ;
      //caribration_flag = false;
    }
    count++; 
}
/* 座標比較の関数 */
void RobotManager::comparePostion(){
    //ROS_INFO("comparePostion");
    /*座標と座標の直線距離を計算*/
    float x1 = robot_camerapose.position.x;
    float y1 = robot_camerapose.position.y;
    float x2 = robot_selfpose.position.x;
    float y2 = robot_selfpose.position.y;
    //ROS_INFO("x1:%f , y1:%f , x2:%f , y2:%f ",x1,y1,x2,y2);
    distance = std::sqrt((x2 - x1) * (x2 - x1) + (y2 - y1) * (y2 - y1));

}
void RobotManager::robotmsgcheck(){
    ROS_INFO("robotmsgcheck");
    if(robot_msg.data == "finish")caribration_flag = false;

}
/* AMCLコールバック関数*/
void RobotManager::amclCallback (const geometry_msgs::PoseWithCovarianceStamped& msg)
{
    //ROS_INFO("amclCallback");
    robot_selfpose = msg.pose.pose;
    amcl_flag=true;
    
}
/* ロボット位置推定コールバック関数*/
void RobotManager::poseCallback(const geometry_msgs::Pose pose)
{
    ROS_INFO("poseCallbackCallback");
    robot_camerapose = pose;
    camerapose_flag=true;
    
}
/* ロボットメッセージコールバック関数*/
void RobotManager::robotmsgCallback(const std_msgs::String message)
{
    //ROS_INFO("robotmsgCallback");
    robot_msg = message;
    robotmsgcheck();
}

void RobotManager::mainloop()
{
    ROS_INFO("mainloop ");
    ros::Rate loop_rate(10);
    while(ros::ok)
    {
        
        ros::spinOnce();
        loop_rate.sleep();
        //robotmsgcheck();
        //ROS_INFO("robot(amcl) x:%f , y;%f",robot_selfpose.position.x, robot_selfpose.position.y);
        //ROS_INFO("camera      x:%f , y;%f",robot_camerapose.position.x, robot_camerapose.position.y);
        //ROS_INFO("caribration_flag=%d ,acml_flag=%d ,camerapose_flag=%d",caribration_flag,amcl_flag,camerapose_flag);

        if(caribration_flag == false && amcl_flag == true && camerapose_flag == true)
        {
 
            /*座標比較の関数*/
            comparePostion();
            if(threshold_distance < distance)
            {
                ROS_INFO("calibration Push");
                ROS_INFO("calib_push threshold_distance %f, distance %f",threshold_distance, distance);
                ROS_INFO("calib_push robot(amcl) x:%f , y;%f",robot_selfpose.position.x, robot_selfpose.position.y);
                ROS_INFO("calib_push camera      x:%f , y;%f",robot_camerapose.position.x, robot_camerapose.position.y);
                ROS_INFO("calib_push camera      x:%f , y;%f",robot_camerapose.position.x, robot_camerapose.position.y);

                /*パブリッシャーでキャリブレーション指示を出力*/
		robotpose_pub.publish(robot_camerapose);
                /*パブリッシャーで座標を出力*/
                std_msgs::String message;
                message.data="test";
		calibration_pub.publish(message);
                /*キャリブレーションフラグをON*/
                caribration_flag = true;
                robot_msg.data ="";
            }
            else
            {
                ROS_INFO("no calibration");
                ROS_INFO("no calib threshold_distance  %f, distance %f",threshold_distance, distance);
                //ROS_INFO("no calib robot(amcl)  x:%f,y;%f",robot_selfpose.position.x, robot_selfpose.position.y);
                //ROS_INFO("no calib camera       x:%f,y;%f",robot_camerapose.position.x, robot_camerapose.position.y);
                //dummy();
            }
            amcl_flag = false; 
            camerapose_flag = false;
        }
        else
        {
            //ROS_INFO("calibration flag=false");
            //ROS_INFO("calib_false robot(amcl) x:%f , y;%f",robot_selfpose.position.x, robot_selfpose.position.y);
            //ROS_INFO("calib_false camera      x:%f , y;%f",robot_camerapose.position.x, robot_camerapose.position.y);
            //dummy();
        }

    }
}

int main (int argc, char** argv)
{
    ros::init (argc, argv, "robotmanager");


    RobotManager robotmanager;
    ROS_INFO("ROS Init");
    robotmanager.setup();
    robotmanager.mainloop();
}

