#include <ros/ros.h>
#include <sensor_msgs/PointCloud2.h>
#include <sensor_msgs/CompressedImage.h>
#include <sensor_msgs/Image.h>
#include <geometry_msgs/PoseArray.h>
// PCL specific includes
#include <pcl_conversions/pcl_conversions.h>
#include <pcl/io/pcd_io.h>
#include <pcl/point_types.h>
//#include <pcl/ros/conversions.h> can not use now
#include <pcl/point_cloud.h>
#include <pcl_ros/point_cloud.h>
// For opencv
#include <opencv2/core.hpp>
#include <opencv2/opencv.hpp>
// OpenCV
#include <cv_bridge/cv_bridge.h>
//
// image transport
#include <sensor_msgs/Image.h>
#include <sensor_msgs/image_encodings.h>
#include <image_transport/image_transport.h>
#include <cv_bridge/cv_bridge.h>
// ARマーカ関連
#include <opencv2/aruco.hpp>
//QR
#include <zbar.h>
//回転の仕方
#include <tf/transform_broadcaster.h>

#include <cmath>
#include <math.h>//ragian
#include <iostream>
#include <string>
#define PI 3.141592653589793

#include <fstream>
#include <string>

using namespace std;

class RobotRecognaize{
    private:
        ros::NodeHandle nh;
        ros::Subscriber cloud_sub;
        ros::Subscriber image_sub;
        ros::Subscriber camerapose_sub ;
        ros::Publisher  robotpose_pub;
        //pcl::PointCloud<pcl::PointXYZRGB>::ptr cloud(new pcl::PointCloud<pcl::PointXYZRGB>);
        pcl::PointCloud<pcl::PointXYZRGB> cloud;
        //pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud{new pcl::PointCloud<pcl::PointXYZRGB>};;
        pcl::PointCloud<pcl::PointXYZRGB> worldcloud;
        
        cv::Mat jpgimage;
        cv::Mat pclimage;
        float pointnum[4][2];//QR 4 corner point x,y
        // QR x,y,theta　=Robot coordinate
        float qrcenterx;
        float qrcentery;
        float qrcentertheta;
        geometry_msgs::Pose camerapose;
        geometry_msgs::Pose robotpose;
        int cameraheight ;
        int camerawidth  ;
        float Rmatrix[9];
        float Cmatrix[9];
        bool imageflag,pclflag,qrrecoflag,outputimage,realsense,pcltoimage;
    public:
        RobotRecognaize();
        void setup(void);
        double getRadian(double x1, double y1, double x2, double y2);
        void getQRlocation(void);
        void Point3DCorrdinate(void);
        geometry_msgs::Quaternion rpy_to_geometry_quat(double roll, double pitch, double yaw);
        void camera_to_world_3Dpoint(void);
        void mainloop(void);
        void Point3DtoOpenCV(void);
        void toRotationMatrix(geometry_msgs::Quaternion q ,float r[9] );
        //callback
        //void cloud_cb(const sensor_msgs::PointCloud2ConstPtr& input);
        void cloud_cb(const sensor_msgs::PointCloud2& input);
        void image_cb (const sensor_msgs::CompressedImage& input);
        void imageCallback(const sensor_msgs::ImageConstPtr& msg);
        void coordinate_cb (const geometry_msgs::Pose& input);

};
RobotRecognaize::RobotRecognaize()
{
    //image_transport::ImageTransport it(nh);
    
    //pubscriber
    //ロボットの座標出力（ロボット認識プログラムからのパブリッシュ）
    robotpose_pub = nh.advertise<geometry_msgs::Pose>("robotpose", 1000);
    //カメラの座標出力（レールコントローラプログラムからのパブリッシュ）
    //cameraposition_pub = nh.advertise<geometry_msgs::Pose>("cameraposition", 1000);
    //sublisher
    cloud_sub = nh.subscribe ("pclpoint2", 10, &RobotRecognaize::cloud_cb,this);
    image_sub = nh.subscribe ("Cameraimage", 10, &RobotRecognaize::image_cb,this);
    //image_sub = nh.subscribe ("Cameraimage", 10, &RobotRecognaize::imageCallback,this);    
    camerapose_sub = nh.subscribe ("camerapose", 10, &RobotRecognaize::coordinate_cb,this);
    //image_sub2 = it.subscribe ("/out_image_base_topic/compressed", 10, &RobotRecognaize::imageCallback,this);    
    
}
void RobotRecognaize::setup()
{
    ros::NodeHandle pnh("~");
    imageflag=false;
    pclflag=false;
    qrrecoflag=false;
    camerapose.position.x = 0;
    camerapose.position.y = 0;
    camerapose.position.z = 0.7;
    camerapose.orientation.x = -0.7071080799;
    camerapose.orientation.y = 0;
    camerapose.orientation.z = 0;
    camerapose.orientation.w = 0.7071054825;
    outputimage = false;
    realsense=false;
    pnh.getParam("outputimage", outputimage);
    pnh.getParam("realsense", realsense);
    pnh.getParam("pcltoimage", pcltoimage);
    //ros::param::get("outputimage", outputimage);
    //ros::param::get("realsense", realsense);
    ROS_INFO("outputimage:%d ,realsense:%d ,pcltoimage:%d",outputimage,realsense,pcltoimage);



}
/*
void RobotRecognaize::cloud_cb (const sensor_msgs::PointCloud2 & input)
{
    //ROS_INFO("input pointcloud");
    pcl::fromROSMsg(input, cloud);
    cameraheight = cloud.height;
    camerawidth  = cloud.width;
    pclflag=true;
}*/
void RobotRecognaize::cloud_cb(const sensor_msgs::PointCloud2 & input)
{
  //ROS_INFO("input pointcloud");
  cloud.clear();
  // Get the field structure of this point cloud
  int pointBytes = input.point_step;
  int offset_x;
  int offset_y;
  int offset_z;
  int offset_rgb;
  for (int f=0; f<input.fields.size(); ++f)
  {
    if (input.fields[f].name == "x")
      offset_x = input.fields[f].offset;
    if (input.fields[f].name == "y")
      offset_y = input.fields[f].offset;
    if (input.fields[f].name == "z")
      offset_z = input.fields[f].offset;
    if (input.fields[f].name == "rgb")
      offset_rgb = input.fields[f].offset;
  }

  // populate point cloud object
  int point_length = input.width * input.height;
  for (int p=0; p<point_length; ++p)
  {
      pcl::PointXYZRGB newPoint;

      newPoint.x = *(float*)(&input.data[0] + (pointBytes*p) + offset_x);
      newPoint.y = *(float*)(&input.data[0] + (pointBytes*p) + offset_y);
      newPoint.z = *(float*)(&input.data[0] + (pointBytes*p) + offset_z);
      int rgb_ = *(int*)(&input.data[0] + (pointBytes*p) + offset_rgb);
      newPoint.r = ((rgb_ >> 16) & 0xff);
      newPoint.g = ((rgb_ >> 8) & 0xff);
      newPoint.b = (rgb_ & 0xff);
      cloud.points.push_back(newPoint);
  }
  cloud.height = input.height;
  cloud.width = input.width;
  pclflag=true;
  //ROS_INFO("pointcloud size: %d", cloud.points.size());
  //pcl::io::savePCDFileASCII ("/home/test/pcd_testdata.pcd" , cloud);
  //ROS_INFO("input pointclou end");
}

void RobotRecognaize::image_cb (const sensor_msgs::CompressedImage& input)
{
    //ROS_INFO("input image");
    // Convert the sensor_msgs/CompressedImage data to OpenCV
    //ROS_INFO("imagedata hedar,format,%f, %f, %f",input.frame_id,input.format);
    jpgimage = cv::imdecode(cv::Mat(input.data),cv::IMREAD_COLOR);
    imageflag=true;
}
void RobotRecognaize::imageCallback(const sensor_msgs::ImageConstPtr& msg) {

    try {
        ROS_INFO("imageCallback OK");
        jpgimage = cv_bridge::toCvCopy(msg, sensor_msgs::image_encodings::BGR8)->image;
        imageflag=true;
    }
    catch (cv_bridge::Exception& e) {
        ROS_INFO("imageCallback ERR");

        //ROS_ERROR("cv_bridge exception: %s", e.what());
    }
}
void RobotRecognaize::coordinate_cb (const geometry_msgs::Pose& input)
{
    //ROS_INFO("input cameracoordinate");
    camerapose = input;
 
}
void RobotRecognaize::toRotationMatrix(geometry_msgs::Quaternion q ,float *r) {
  ROS_INFO("toRotationMatrix");
  auto xy2 = q.x * q.y * 2;
  auto xz2 = q.x * q.z * 2;
  auto xw2 = q.x * q.w * 2;
  auto yz2 = q.y * q.z * 2;
  auto yw2 = q.y * q.w * 2;
  auto zw2 = q.z * q.w * 2;
  auto ww2 = q.w * q.w * 2;
  r[0] = ww2 + 2 * q.x * q.x - 1 ;
  r[1] = xy2 + zw2 ;
  r[2] = xz2 - yw2 ;
  r[3] = xy2 - zw2 ;
  r[4] = ww2 + 2 * q.y * q.y - 1 ; 
  r[5] = yz2 + xw2 ;
  r[6] =  xz2 + yw2 ; 
  r[7] = yz2 - xw2 ;
  r[8] = ww2 + 2 * q.z * q.z - 1 ; 
 
}
void RobotRecognaize::camera_to_world_3Dpoint(){
    ROS_INFO("camera_to_world_3Dpoint");
    worldcloud.width = cloud.width; 
    worldcloud.height = cloud.height;
    worldcloud.points.resize(worldcloud.width * worldcloud.height);
    //ROS_INFO("worldcloud %d",worldcloud.points.size());
    toRotationMatrix(camerapose.orientation,Cmatrix);
    //std::ofstream writing_file;
    //std::string filename = "/home/test/choreonoidrengecamera.txt";
    //writing_file.open(filename, std::ios::out);
    for(int i = 0; i < cloud.points.size() ; i++)
    {
        //writing_file <<i<<","<<cloud.points[i].x <<","<<cloud.points[i].y<<","<<cloud.points[i].z<< std::endl;

        //worldcloud[i].x = cloud.points[i].x;
        //worldcloud[i].y = cloud.points[i].y;
        //worldcloud[i].z = cloud.points[i].z;
        // https://www.notion.so/6a9cf2e1dcc24dd5a0ca35dd9ed58297
        // Camera coordinates to World coordinates
        worldcloud[i].x = ( Cmatrix[0]*cloud.points[i].x + Cmatrix[1]*cloud.points[i].z + Cmatrix[2]*cloud.points[i].y ) + camerapose.position.x;
        worldcloud[i].y = ( Cmatrix[3]*cloud.points[i].x + Cmatrix[4]*cloud.points[i].z + Cmatrix[5]*cloud.points[i].y ) + camerapose.position.y;
        worldcloud[i].z = ( Cmatrix[6]*cloud.points[i].x + Cmatrix[7]*cloud.points[i].z + Cmatrix[8]*cloud.points[i].y ) + camerapose.position.z;

    }
    ROS_INFO("camera_to_world_3Dpoint end");
    //writing_file.close();

}

double RobotRecognaize::getRadian(double x1, double y1, double x2, double y2) {
    double radian = atan2(y2 - y1,x2 - x1);
    return radian;
}

void RobotRecognaize::getQRlocation(){
    cv::Mat grey;
    cv::Mat qrimage;
    if(pcltoimage==true)
        qrimage=pclimage;
    else
        qrimage=jpgimage;
        
    int imagewidth = qrimage.cols;  
    int imageheight = qrimage.rows;  
    int qrx=0;
    int qry=0;
    qrrecoflag=false;
    cv::cvtColor(qrimage,grey,cv::COLOR_BGR2GRAY);
    uchar *raw = (uchar *)grey.data;  
    // wrap image data  
    zbar::Image image2(imagewidth, imageheight, "Y800", raw, imagewidth * imageheight);  
    // scan the image for barcodes  
    zbar::ImageScanner scanner;  
    scanner.set_config(zbar::ZBAR_NONE, zbar::ZBAR_CFG_ENABLE, 1);  
    int QRfindnum = scanner.scan(image2);
    // extract results  
    for(zbar::Image::SymbolIterator symbol = image2.symbol_begin();  
        symbol != image2.symbol_end(); ++symbol) {  
        //vector<cv::Point> vp;  
        int n = symbol->get_location_size();  
        cout<<"qr data"<<symbol->get_data()<<endl;
        for(int i=0;i<n;i++){  
            // QR center coordinates(pixel)
            qrx += symbol->get_location_x(i)/n;
            qry += symbol->get_location_y(i)/n;
            pointnum[i][0]=symbol->get_location_x(i);
            pointnum[i][1]=symbol->get_location_y(i);

        }
        qrrecoflag=true;
        qrcenterx=qrx;
        qrcentery=qry;
    }
    if(qrrecoflag==true)
        ROS_INFO("qrcode center x y: %f , %f",qrcenterx,qrcentery);
}
void RobotRecognaize::Point3DCorrdinate(){
    //camera_to_world_3Dpoint();
    // robot x , y , z
    toRotationMatrix(camerapose.orientation,Cmatrix);
    int cloudnum = qrcentery * camerawidth + qrcenterx;
    //robotpose.position.x = worldcloud[qrcentery * camerawidth + qrcenterx].x ;
    //robotpose.position.y = worldcloud[qrcentery * camerawidth + qrcenterx].y ;
    //robotpose.position.z = worldcloud[qrcentery * camerawidth + qrcenterx].z ;
    ROS_INFO("camera x y z: %f , %f ,%f",camerapose.position.x,camerapose.position.y,camerapose.position.z);
    ROS_INFO("matrix 0: %f , %f , %f",Cmatrix[0],Cmatrix[1],Cmatrix[2]);
    ROS_INFO("matrix 3: %f , %f , %f",Cmatrix[3],Cmatrix[4],Cmatrix[5]);
    ROS_INFO("matrix 6: %f , %f , %f",Cmatrix[6],Cmatrix[7],Cmatrix[8]);

    ROS_INFO("cloud x y z : %f , %f , %f",cloud.points[cloudnum].x,cloud.points[cloudnum].y,cloud.points[cloudnum].z);

    //pointcloud xyz to choreonoid xyz 
    int realsenseytoward =-1;
    int realsenseztoward =-1;
    if(realsense ==false){
        realsenseytoward=1;
        realsenseztoward=-1;
    }
    double worldx = cloud.points[cloudnum].x;
    double worldy = cloud.points[cloudnum].z * realsenseztoward;
    double worldz = cloud.points[cloudnum].y * realsenseytoward;

    robotpose.position.x= ( Cmatrix[0]*worldx + Cmatrix[3]*worldy + Cmatrix[6]*worldz ) + camerapose.position.x;
    robotpose.position.y= ( Cmatrix[1]*worldx + Cmatrix[4]*worldy + Cmatrix[7]*worldz ) + camerapose.position.y;
    robotpose.position.z= ( Cmatrix[2]*worldx + Cmatrix[5]*worldy + Cmatrix[8]*worldz ) + camerapose.position.z;

    // robot degree
    int point1 = pointnum[0][1] * camerawidth + pointnum[0][0];
    int point2 = pointnum[3][1] * camerawidth + pointnum[3][0];
    //QR corners coordinates(world) 
    double qrx1 = cloud.points[point1].x ;
    double qry1 = cloud.points[point1].z * realsenseztoward;
    double qrz1 = cloud.points[point1].y * realsenseytoward;
    double qrx2 = cloud.points[point2].x ;
    double qry2 = cloud.points[point2].z * realsenseztoward;
    double qrz2 = cloud.points[point2].y * realsenseytoward;

    double x1 = ( Cmatrix[0]*qrx1 + Cmatrix[3]*qry1 + Cmatrix[6]*qrz1 ) + camerapose.position.x;
    double y1 = ( Cmatrix[1]*qrx1 + Cmatrix[4]*qry1 + Cmatrix[7]*qrz1 ) + camerapose.position.y;

    double x2 = ( Cmatrix[0]*qrx2 + Cmatrix[3]*qry2 + Cmatrix[6]*qrz2 ) + camerapose.position.x;
    double y2 = ( Cmatrix[1]*qrx2 + Cmatrix[4]*qry2 + Cmatrix[7]*qrz2 ) + camerapose.position.y;
    // get radian from 2 points(x,y)
    double QRradian = getRadian(x1, y1, x2, y2); 
    double QRdegree = QRradian* 180.0 / PI;
    ROS_INFO("robot x y z degree: %f , %f ,%f ,%f",robotpose.position.x,robotpose.position.y,robotpose.position.z,QRdegree);
    robotpose.orientation = rpy_to_geometry_quat(0,0, QRradian);

}

geometry_msgs::Quaternion RobotRecognaize::rpy_to_geometry_quat(double roll, double pitch, double yaw){
	tf::Quaternion quat=tf::createQuaternionFromRPY(roll,pitch,yaw);
	geometry_msgs::Quaternion geometry_quat;
	quaternionTFToMsg(quat, geometry_quat);
	return geometry_quat;
}

void RobotRecognaize::Point3DtoOpenCV() {
    //ROS_INFO("Point3DtoOpenCV");
    //OpenCV x-yImage 
    cameraheight = cloud.height;
    camerawidth  = cloud.width;
    pclimage = cv::Mat(cameraheight, camerawidth, CV_8UC3); 
    for(int y=0;y<pclimage.rows;y++) 
    { 
        for(int x=0;x<pclimage.cols;x++) 
        { 
             cv::Vec3b color = cv::Vec3b(
                               cloud.points.at(y*pclimage.cols+x).b,
                               cloud.points.at(y*pclimage.cols+x).g,
                               cloud.points.at(y*pclimage.cols+x).r); 

             pclimage.at<cv::Vec3b>(cv::Point(x,y)) = color; 
           } 
    }
   //cv::imwrite("/home/test/pclimage.png",pclimage);
}

void RobotRecognaize::mainloop(){

    ros::Rate rate(100.0);
    //Recognaize Mark
    while(ros::ok())
    {
      ros::spinOnce();
      rate.sleep();
      if(imageflag==true && pclflag==true){
          /*if(outputimage==true){
              Point3DtoOpenCV();
              cv::imwrite("/home/test/jpgimage.png",jpgimage);
          }*/
          Point3DtoOpenCV();
          getQRlocation();
          if(qrrecoflag==true){
              Point3DCorrdinate();
              robotpose_pub.publish(robotpose);
          }
          imageflag = false;
          pclflag = false;
      }

    }

}

int main (int argc, char** argv)
{
    // Initialize ROS
    ros::init (argc, argv, "robotrecognize");
    ROS_INFO("ROS Init");
    RobotRecognaize robot_rec;
    robot_rec.setup();
    robot_rec.mainloop();
}
